# -*- coding: utf-8 -*-
#
#    Error Tracking app
#
#    :copyright: 2020 Sonu Kumar
#    :license: BSD-3-Clause
#

__version__ = '1.1.6'
__author__ = 'Sonu Kumar'
__email__ = 'sonunitw12@gmail.com'

from error_tracker.libs.mixins import *
from error_tracker.flask import *
from error_tracker.django import *
from error_tracker.flask.utils import configure_scope as flask_scope
from error_tracker.django.apps import DjangoErrorTracker
from error_tracker.django.utils import capture_message, track_exception, configure_scope, capture_exception
from error_tracker.libs.exception_formatter import *

__all__ = [
    # flask modules
    "AppErrorTracker", "DefaultFlaskContextBuilder", "flask_scope",

    # mixin classes
    "NotificationMixin", "ModelMixin", "MaskingMixin",
    "ContextBuilderMixin", "TicketingMixin", "ViewPermissionMixin",

    # Django modules
    "DefaultDjangoContextBuilder", "DjangoErrorTracker", "DefaultDjangoViewPermission",
    "capture_message", "track_exception", "configure_scope", "capture_exception",

    # lower level methods
    "format_exception", "print_exception"
]



# -*- coding: utf-8 -*-
#
#    Flask components
#
#    :copyright: 2019 Sonu Kumar
#    :license: BSD-3-Clause
#


from .flask_error import AppErrorTracker
from .utils import DefaultFlaskContextBuilder

__all__ = ["AppErrorTracker", "DefaultFlaskContextBuilder"]



# -*- coding: utf-8 -*-
#
#    Django error tracker default model
#
#    :copyright: 2019 Sonu Kumar
#    :license: BSD-3-Clause
#
from django.core.paginator import Paginator, EmptyPage
from django.db import models
from error_tracker.libs.mixins import ModelMixin
from django.utils.timezone import now
from traceback import print_exc
from .settings import EXCEPTION_APP_DEFAULT_LIST_SIZE
from collections import namedtuple

Page = namedtuple("Page", "has_next, next_num, has_prev, prev_num, items ")


class ErrorModel(models.Model, ModelMixin):
    """
    Model to track exceptions
    """
    hash = models.CharField(max_length=64, primary_key=True)
    host = models.CharField(max_length=1024)
    path = models.CharField(max_length=4096)
    method = models.CharField(max_length=64)
    request_data = models.TextField()
    exception_name = models.CharField(max_length=256)
    traceback = models.TextField()
    count = models.IntegerField(default=0)
    created_on = models.DateTimeField(auto_now=True)
    last_seen = models.DateTimeField(auto_now=True, db_index=True)

    @classmethod
    def get_exceptions_per_page(cls, page_number=1):
        records = cls.objects.all().order_by('last_seen')
        paginator = Paginator(records, EXCEPTION_APP_DEFAULT_LIST_SIZE)
        try:
            page = paginator.page(page_number)
            return Page(page.has_next(),
                        page.next_page_number() if page.has_next() else None,
                        page.has_previous(),
                        page.previous_page_number() if page.has_previous() else None,
                        page.object_list)
        except EmptyPage:
            return Page(False, None, True, paginator.num_pages, [])

    @classmethod
    def get_entity(cls, rhash):
        return cls.objects.get(pk=rhash)

    @classmethod
    def create_or_update_entity(cls, rhash, host, path, method, request_data, exception_name, traceback):
        try:
            obj, created = cls.objects.get_or_create(hash=rhash)
            if created:
                obj.host, obj.path, obj.method, obj.request_data, obj.exception_name, obj.traceback = \
                    host, path, method, request_data, exception_name, traceback
                obj.count = 1
                obj.save()
            else:
                obj.count += 1
                obj.last_seen = now()
                obj.save(update_fields=['count', 'last_seen'])
        except Exception:
            print_exc()

    @classmethod
    def delete_entity(cls, rhash):
        return cls.objects.filter(pk=rhash).delete()

    class Meta:
        db_table = 'exceptions'
        swappable = 'APP_ERROR_DB_MODEL'



# -*- coding: utf-8 -*-
#
#    Django error tracker app config
#
#    :copyright: 2019 Sonu Kumar
#    :license: BSD-3-Clause
#


from django.apps import AppConfig
from django.utils.translation import gettext_lazy as _


class DjangoErrorTracker(AppConfig):
    name = 'error_tracker.django'
    label = 'error_tracker'
    verbose_name = _("Error Monitoring and Exception Tracking")



# -*- coding: utf-8 -*-
#
#    Django error tracker utils classes
#
#    :copyright: 2020 Sonu Kumar
#    :license: BSD-3-Clause
#

import re
import json

from django.http import RawPostDataException

from error_tracker.libs.mixins import ContextBuilderMixin, NotificationMixin, ViewPermissionMixin
from error_tracker.libs.utils import get_context_dict
from django.core.mail import send_mail


class DefaultDjangoContextBuilder(ContextBuilderMixin):
    """
    Default request builder, this records, form data, header and URL parameters and mask them if necessary
    """

    @staticmethod
    def _get_form_data(request):
        form = {}
        if request is None:
            return form
        post = request.POST
        if post is None or len(post) == 0:
            body = None
            try:
                body = request.data
            except AttributeError:
                try:
                    body = request.body
                except RawPostDataException:
                    pass
            if body is not None:
                try:
                    from rest_framework.request import Request
                    if isinstance(body, Request):
                        form = body.data
                except ImportError:
                    pass
                if len(form) == 0 and len(body) > 0:
                    try:
                        form = json.loads(body, encoding="UTF-8")
                    except Exception:
                        form = {'data': body}
        else:
            form = post.dict()
        return form

    @staticmethod
    def _get_headers(request):
        if request is not None:
            try:
                headers = request.headers.dict()
            except AttributeError:
                regex = re.compile('^HTTP_')
                headers = dict((regex.sub('', header), value) for (header, value)
                               in request.META.items() if header.startswith('HTTP_'))
            return headers

    @staticmethod
    def _get_args(request):
        if request is not None:
            return request.GET.dict()

    def get_context(self, request, masking=None, additional_context=None):
        return str(get_context_dict(headers=self._get_headers(request),
                                    form=self._get_form_data(request),
                                    args=self._get_args(request),
                                    context=additional_context,
                                    masking=masking))


class DjangoNotification(NotificationMixin):
    """
    Send emails to the configured users
    """

    def notify(self, request, exception,
               email_subject=None,
               email_body=None,
               from_email=None,
               recipient_list=None):
        if recipient_list is not None and from_email is not None:
            send_mail(email_subject, email_body, from_email, recipient_list, fail_silently=True)


class DefaultDjangoViewPermission(ViewPermissionMixin):

    def __call__(self, request):
        if hasattr(request.user, "is_superuser"):
            return request.user.is_superuser
        return False


class configure_scope(object):
    """
    Use this class to work with context manager where more context can be added on the fly

    usage
    with configure_scope(request=request) as scope:
        scope.set_extra("id", 1234)

    """

    def __init__(self, request=None, handle_exception=True, context=None):
        """
        :param request:  current request object
        :param handle_exception:  whether exception has to be re-raised or not
        :param context: initial context detail
        """
        self.context = context or {}
        self.request = request
        self.handle_exception = handle_exception

    def set_extra(self, key, value):
        """
        Add key value in context detail
        :param key:  context key
        :param value: context value
        :return:
        """
        self.context[key] = value

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if exc_type is not None:
            capture_exception(request=self.request,
                              exception=exc_type,
                              additional_context=self.context)
        return self.handle_exception


def capture_message(message, exception=None, request=None):
    """
    Use this method to capture any message once exception is excepted
    :param message:  message to be recorded
    :param exception:  exception occurred
    :param request:  current request object
    :return:  None
    """
    return capture_exception(request=request, exception=exception,
                             additional_context={'message': message})


def track_exception(func, additional_context=None, request=None, silent=False):
    """
    Decorator to be used for automatic exception capture
    :param func:  function on which it has been used
    :param additional_context:  additional context detail
    :param request:  current request
    :param silent: exception should be re-raised or not
    :return: None
    """

    def wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except Exception as ex:
            capture_exception(request=request, exception=ex, additional_context=additional_context)
            if not silent:
                raise ex

    return wrapper


def capture_exception(request=None, exception=None, additional_context=None):
    """
    Use this method to capture any exception after it has been captured using try except
    :param exception:  exception occurred
    :param request:  current request object
    :param additional_context: any additional context detail
    :return:  None
    """
    from error_tracker.django.middleware import error_tracker
    error_tracker.capture_exception(request=request, exception=exception,
                                    additional_context=additional_context)



# -*- coding: utf-8 -*-
#
#    Django error tracker default settings
#
#    :copyright: 2020 Sonu Kumar
#    :license: BSD-3-Clause
#

from django.conf import settings as dj_settings
from django.core.exceptions import ImproperlyConfigured


def get(key, default):
    try:
        return getattr(dj_settings, key, default)
    except ImproperlyConfigured:
        return default


# App's HOME page default page size
EXCEPTION_APP_DEFAULT_LIST_SIZE = get('EXCEPTION_APP_DEFAULT_LIST_SIZE', 20)
# what all sensitive data should be masked, this means any variables whose name have
# either password or secret would be masked
APP_ERROR_MASKED_KEY_HAS = get('APP_ERROR_MASKED_KEY_HAS', ("password", "secret"))
# Sensitive data masking value
APP_ERROR_MASK_WITH = get('APP_ERROR_MASK_WITH', '*************')
# exception email subject prefix
APP_ERROR_SUBJECT_PREFIX = get('APP_ERROR_SUBJECT_PREFIX', '')
# whose email would be used to send email
APP_ERROR_EMAIL_SENDER = get('APP_ERROR_EMAIL_SENDER', None)
# whom email should be send in the case of failure
APP_ERROR_RECIPIENT_EMAIL = get('APP_ERROR_RECIPIENT_EMAIL', None)
# Whether all types of errors should be tracked or not e.g. 404, 401 etc
TRACK_ALL_EXCEPTIONS = get('TRACK_ALL_EXCEPTIONS', False)
# Path to masking module
APP_ERROR_MASKING_MODULE = get('APP_ERROR_MASKING_MODULE', None)
# path to ticketing module, it should not have any constructor parameters
APP_ERROR_TICKETING_MODULE = get('APP_ERROR_TICKETING_MODULE', None)
# path to notifier module, that would send notifications to entities
APP_ERROR_NOTIFICATION_MODULE = get('APP_ERROR_NOTIFICATION_MODULE', None)
# path to custom context builder
APP_ERROR_CONTEXT_BUILDER_MODULE = get('APP_ERROR_CONTEXT_BUILDER_MODULE', None)
# In case of different database model, provide path to that
APP_ERROR_DB_MODEL = get('APP_ERROR_DB_MODEL', None)
# Check error views are visible to others or not
APP_ERROR_VIEW_PERMISSION = get('APP_ERROR_VIEW_PERMISSION', None)



# -*- coding: utf-8 -*-
#
#    Error tracker flasks plugins default value
#
#    :copyright: 2019 Sonu Kumar
#    :license: BSD-3-Clause
#

# Whether notification should be send or not
APP_ERROR_SEND_NOTIFICATION = False
# List of email recipients, these users would be send an email
APP_ERROR_RECIPIENT_EMAIL = None
# Error email subject prefix
APP_ERROR_SUBJECT_PREFIX = ""
# Sensitive data masking value
APP_ERROR_MASK_WITH = "**************"
# what all sensitive data should be masked, this means any variables whose name have
# either password or secret would be masked
APP_ERROR_MASKED_KEY_HAS = ("password", "secret")
# APP URL prefix where endpoint would be exposed
APP_ERROR_URL_PREFIX = "/dev/error"
# Email sender user email's ID
APP_ERROR_EMAIL_SENDER = None
# In Exceptions listing page, number of entry should be displayed
APP_DEFAULT_LIST_SIZE = 20



# -*- coding: utf-8 -*-
#
#    Django error tracker default value
#
#    :copyright: 2020 Sonu Kumar
#    :license: BSD-3-Clause
#
from django.http import Http404, HttpResponse
from django.shortcuts import redirect, render
from django.urls import reverse
from django.views.decorators.http import require_GET
from error_tracker.django import get_exception_model, get_view_permission

model = get_exception_model()

view_permission = get_view_permission()


def has_view_permission(func):
    def wrapper(request, *args, **kwargs):
        if view_permission(request):
            return func(request, *args, **kwargs)
        return HttpResponse(status=401)

    return wrapper


@require_GET
@has_view_permission
def view_list(request):
    """
    Home page that lists mose recent exceptions
    :param request: request object
    :return:  rendered template
    """
    title = "App Error"
    try:
        page = int(request.GET.get('page', 1))
    except:
        page = 1
    error = False
    errors = model.get_exceptions_per_page(page_number=page)
    next_url = reverse('view_errors') + "?page=" + str(errors.next_num) \
        if errors.has_next else None
    prev_url = reverse('view_errors') + "?page=" + str(errors.prev_num) \
        if errors.has_prev else None

    return render(request, template_name='error_tracker/list.html',
                  context=dict(error=error, title=title, errors=errors,
                               next_url=next_url, prev_url=prev_url))


@require_GET
@has_view_permission
def delete_exception(request, rhash):
    """
    Delete an exceptions
    :param request:  request object
    :param rhash:  hash key of the exception
    :return: redirect back to home page
    """
    model.delete_entity(rhash)
    return redirect(reverse('view_errors'))


@require_GET
@has_view_permission
def detail(request, rhash):
    """
    Display a specific page of the exception
    :param request: request object
    :param rhash:  hash key of the exception
    :return: detailed view
    """
    obj = model.get_entity(rhash)
    error = False
    if obj is None:
        raise Http404
    title = "%s : %s" % (obj.method, obj.path)
    return render(request, template_name='error_tracker/detail.html',
                  context=dict(error=error, title=title, obj=obj))



# -*- coding: utf-8 -*-
#
#    Django error tracker middleware responsible for recording exception
#
#    :copyright: 2020 Sonu Kumar
#    :license: BSD-3-Clause
#

from error_tracker.django import get_masking_module, get_context_builder, get_ticketing_module, \
    get_exception_model, get_notification_module, APP_ERROR_SUBJECT_PREFIX, APP_ERROR_EMAIL_SENDER, \
    APP_ERROR_RECIPIENT_EMAIL, TRACK_ALL_EXCEPTIONS
from error_tracker.libs.utils import get_exception_name, get_context_detail, get_notification_subject

model = get_exception_model()
ticketing = get_ticketing_module()
masking = get_masking_module()
notifier = get_notification_module()
context_builder = get_context_builder()


# noinspection PyMethodMayBeStatic
class ErrorTracker(object):
    """
     ErrorTracker class, this is responsible for capturing exceptions and
     sending notifications and taking other actions,
    """

    @staticmethod
    def _send_notification(request, message, exception, error):
        """
        Send notification to the list of entities or call the specific methods
        :param request: request object
        :param message: message having frame details
        :param exception: exception that's triggered
        :param error:  error model object
        :return: None
        """
        if notifier is None:
            return
        if request is not None:
            method = request.method
            url = request.get_full_path()
        else:
            method = ""
            url = ""
        subject = get_notification_subject(APP_ERROR_SUBJECT_PREFIX,
                                           method, url, exception)
        notifier.notify(request,
                        error,
                        email_subject=subject,
                        email_body=message,
                        from_email=APP_ERROR_EMAIL_SENDER,
                        recipient_list=APP_ERROR_RECIPIENT_EMAIL)

    @staticmethod
    def _raise_ticket(request, error):
        if ticketing is None:
            return
        ticketing.raise_ticket(request, error)

    @staticmethod
    def _post_process(request, frame_str, frames, error):
        if request is not None:
            message = ('URL: %s' % request.path) + '\n\n'
        else:
            message = ""
        message += frame_str
        ErrorTracker._send_notification(request, message, frames[-1][:-1], error)
        ErrorTracker._raise_ticket(request, error)

    def capture_exception(self, request=None, exception=None, additional_context=None):
        """
        Record the exception details and do post processing actions. this method can be used to track any exceptions,
        even those are being excepted using try/except block.
        :param request:  request object
        :param exception: what type of exception has occurred
        :param additional_context: any additional context
        :return:  None
        """
        if request is not None:
            path = request.path
            host = request.META.get('HTTP_HOST', '')
            method = request.method
        else:
            path = ""
            host = ""
            method = ""

        ty, frames, frame_str, traceback_str, rhash, request_data = \
            get_context_detail(request, masking, context_builder,
                               additional_context=additional_context)
        error = model.create_or_update_entity(rhash, host, path, method,
                                              str(request_data),
                                              get_exception_name(ty),
                                              traceback_str)
        ErrorTracker._post_process(request, frame_str, frames, error)


class ExceptionTrackerMiddleWare(ErrorTracker):
    """
    Error tracker middleware that's invoked in the case of exception occurs,
    this should be placed at the end of Middleware lists
    """

    def __init__(self, get_response):
        self.get_response = get_response

    def __call__(self, request):
        return self.get_response(request)

    def process_exception(self, request, exception):
        if exception is None and not TRACK_ALL_EXCEPTIONS:
            return
        self.capture_exception(request, exception)


# use this object to track errors in the case of custom failures, where try/except is used
error_tracker = ErrorTracker()



# -*- coding: utf-8 -*-
#
#    Error tracker's flask plugin, this class initialize it's internal state
#
#    :copyright: 2020 Sonu Kumar
#    :license: BSD-3-Clause
#


from warnings import warn

try:
    from flask import _app_ctx_stack as stack
except ImportError:
    from flask import _request_ctx_stack as stack
from flask import request

from error_tracker.libs.utils import Masking, ConfigError, get_exception_name, \
    get_context_detail, get_notification_subject
from error_tracker.libs.mixins import ModelMixin
from error_tracker.flask.utils import DefaultFlaskContextBuilder, DefaultFlaskViewPermission
from .view import Views
from . import defaults
import datetime

from sqlalchemy import Column, String, Text, Integer, DateTime, desc
from sqlalchemy.orm.exc import NoResultFound

page_size = None


class AppErrorTracker(object):
    def __init__(self, app=None, db=None, model=None, notifier=None, recipients=None,
                 db_table_name="app_error", notification_subject_prefix=None,
                 url_prefix=None, masking=None, ticketing=None,
                 context_builder=DefaultFlaskContextBuilder(),
                 view_permission=DefaultFlaskViewPermission()):
        """
        An error tracker class, which manage the exception and store them to database
        :param app: a Flask app instance where exception has to be tracked
        :param db: Database connection object for storing exception
        :param model: Database model object for tracking exception
        :param notifier: a notifier object that would notify notification to the subscribers
        :param recipients: list of recipients emails
        :param db_table_name: database table name in case if it's required overriding
        :param notification_subject_prefix: notification subject prefix
        :param url_prefix: URL prefix to be exposed by the web apps
        :param masking: a masking object or lambda function that can provide custom masking rules
        :param context_builder: a builder function that provides the request details,
        it can be used to provide custom data, apart from the default one
        :param ticketing:  a ticking mixing object used to create or update ticket in ticketing system
        """
        # Database related fields
        self.app = app
        self.db = db
        self.db_table = db_table_name
        self.model = model
        # masking related fields
        self.masking = masking
        # context builder object
        self.context_builder = context_builder
        self.views = None

        # notification related fields
        self.notifier = notifier
        self.notification_subject_prefix = notification_subject_prefix
        self.send_notification = False
        self.recipients = recipients
        self.notification_sender = None
        self.ticketing = ticketing
        self.view_permission = view_permission

        self.active = False
        if self.app:
            self.init_app(app, db, model=model, notifier=notifier, url_prefix=url_prefix,
                          masking=masking, context_builder=context_builder,
                          ticketing=ticketing, view_permission=view_permission)

    def init_app(self, app, db, model=None, notifier=None, url_prefix=None, masking=None,
                 context_builder=DefaultFlaskContextBuilder(),
                 ticketing=None, view_permission=DefaultFlaskViewPermission()):
        """
        Initialize this app with different attributes
        :param view_permission: view permission checker
        :param app: a Flask app instance
        :param db: Database connection object
        :param model: Database model class
        :param notifier: notifier app which will be used to notify notification
        :param url_prefix: Url prefix for WEB UI
        :param masking: a masking interface object or lambda function
        :param context_builder: a builder function that provides the request details,
        it can be used to provide custom data, apart from the default one
        :param ticketing :   a ticking mixing object used to create or update ticket in ticketing system
        :return: None
        """
        if self.active:
            raise ConfigError("App is already configured")
        self.view_permission = view_permission
        self.app = app
        self.db = db
        if app is None:
            raise ConfigError("app is None")
        if model is None and self.db is None:
            raise ConfigError("Either db or model must be provide")

        send_notification = self.app.config.setdefault('APP_ERROR_SEND_NOTIFICATION',
                                                       defaults.APP_ERROR_SEND_NOTIFICATION)
        recipients = self.app.config.setdefault('APP_ERROR_RECIPIENT_EMAIL',
                                                defaults.APP_ERROR_RECIPIENT_EMAIL)
        subject_prefix = self.app.config.setdefault('APP_ERROR_SUBJECT_PREFIX',
                                                    defaults.APP_ERROR_SUBJECT_PREFIX)
        mask_with = self.app.config.setdefault('APP_ERROR_MASK_WITH', defaults.APP_ERROR_MASK_WITH)
        mask_key_has = self.app.config.setdefault('APP_ERROR_MASKED_KEY_HAS',
                                                  defaults.APP_ERROR_MASKED_KEY_HAS)
        url_prefix = url_prefix or self.app.config.setdefault('APP_ERROR_URL_PREFIX',
                                                              defaults.APP_ERROR_URL_PREFIX)
        self.notification_sender = self.app.config.setdefault('APP_ERROR_EMAIL_SENDER',
                                                              defaults.APP_ERROR_EMAIL_SENDER)

        global page_size
        self.model = model or self._get_model()
        self.views = Views(self.app, self.model, url_prefix, self.view_permission)
        page_size = self.app.config.setdefault("APP_DEFAULT_LIST_SIZE", defaults.APP_DEFAULT_LIST_SIZE)
        # masking object setting
        if type(mask_key_has) == str:
            mask_key_has = (mask_key_has,)
        if mask_key_has and mask_with:
            self.masking = masking or Masking(mask_with, mask_key_has)

        self.notifier = notifier or getattr(app, "notifier", None) or getattr(app, "mailer", None)
        self._set_notification_fields(recipients, send_notification, subject_prefix)
        # Use the new style teardown_appcontext if it's available,
        # otherwise fall back to the request context
        if hasattr(app, 'teardown_appcontext'):
            app.teardown_appcontext(self.teardown)
        else:
            app.teardown_request(self.teardown)
        self.context_builder = context_builder
        self.ticketing = ticketing
        self.active = True

    def _set_notification_fields(self, recipients, send_notification, subject_prefix):
        if self.notifier is not None and send_notification:
            if recipients not in [None, ""]:
                self.send_notification = True
                self.recipients = recipients
                if type(self.recipients) == str:
                    self.recipients = [self.recipients]
                if self.notification_sender is None:
                    raise ConfigError(
                        "Email recipients is set but notification sender is"
                        " not configured. set APP_ERROR_EMAIL_SENDER in app config")
                self.recipients = list(self.recipients)
            else:
                warn("APP_ERROR_RECIPIENT_EMAIL is not set in the app config")

            if subject_prefix in [None, ""]:
                warn("APP_ERROR_SUBJECT_PREFIX is not set in the app config")
            else:
                self.notification_subject_prefix = subject_prefix

    def teardown(self, exception):
        pass

    def _get_model(self):
        """
        :return: Default model for storing exception
        """

        class AppDbModel(ModelMixin, self.db.Model):
            db = self.db
            hash = Column(String(64), primary_key=True)
            host = Column(String(2048))
            path = Column(String(512))
            method = Column(String(32))
            request_data = Column(Text)
            exception_name = Column(String(256))
            traceback = Column(Text)
            count = Column(Integer, default=1)
            created_on = Column(DateTime)
            last_seen = Column(DateTime)

            def __repr__(self):
                return "AppDbModel(%s)" % self.__str__()

            @classmethod
            def delete_entity(cls, rhash):
                """
                :param rhash: lookup key
                :return: None
                """
                cls.query.filter_by(hash=rhash).delete()
                cls.db.session.commit()

            @classmethod
            def create_or_update_entity(cls, rhash, host, path, method, request_data,
                                        exception_name, traceback):
                try:
                    error = cls.query.filter_by(hash=rhash).one()
                    error.count += 1
                    error.last_seen = datetime.datetime.now()
                    # error.exception = exception
                    cls.db.session.commit()
                except NoResultFound:
                    now = datetime.datetime.now()
                    error = cls(hash=rhash, host=host, path=path, method=method,
                                request_data=str(request_data),
                                exception_name=exception_name, traceback=traceback,
                                created_on=now,
                                last_seen=now)
                    cls.db.session.add(error)
                    cls.db.session.commit()
                return error

            @classmethod
            def get_exceptions_per_page(cls, page_number=1):
                return cls.query.order_by(desc(cls.last_seen)).paginate(
                    page_number, page_size, False)

            @classmethod
            def get_entity(cls, rhash):
                """
                :param rhash: key for lookup
                :return: Single entry of this class
                """
                error = cls.query.filter_by(hash=rhash).first()
                return error

            class Meta:
                table_name = self.db_table

        return AppDbModel

    def _send_notification(self, url, method, message, exception, error):
        subject = get_notification_subject(self.notification_subject_prefix,
                                           method, url, exception)
        self.notifier.notify(request, error, email_subject=subject,
                             email_body=message, from_email=self.notification_sender,
                             recipient_list=self.recipients)

    def _post_process(self, rq, frame_str, frames, error):
        if self.send_notification:
            if rq:
                message = ('URL: %s' % rq.path) + '\n\n'
                method = rq.method
                url = rq.url
            else:
                message = ""
                method = ""
                url = ""
            message += frame_str
            self._send_notification(url, method, message,
                                    frames[-1][:-1], error)
        if self.ticketing is not None:
            self.ticketing.raise_ticket(error)

    def capture_message(self, message):
        """
        Capture an error for the current error
        :param message:  message to be recorded
        :return:  None
        """
        self.capture_exception(additional_context={'message': message})

    def capture_exception(self, additional_context=None):
        """
        Record occurred exception, check whether the exception recording is
        enabled or not. If it's enabled then record all the details and store
        them in the database, it will notify email as well. The id of exception
        is SHA256 of frame string
        :return: None
        """
        rq = request
        if not self.active:
            return
        try:
            path = request.path
            host = request.host_url
            method = request.method
        except RuntimeError:
            rq = None
            path = ""
            host = ""
            method = ""

        ty, frames, frame_str, traceback_str, rhash, request_data = \
            get_context_detail(rq, self.masking, self.context_builder, additional_context)
        error = self.model.create_or_update_entity(rhash, host, path, method,
                                                   str(request_data),
                                                   get_exception_name(ty),
                                                   traceback_str)
        self._post_process(rq, frame_str, frames, error)

    def auto_track_exception(self, func, additional_context=None, silent=False):
        """
        Decorator to be used for automatic exception capture, where exception can occur
        :param func:
        :param additional_context:  any additional context
        :param silent: exception should be re-raise or ignored
        :return: None
        """

        def wrapper(*args, **kwargs):
            try:
                return func(*args, **kwargs)
            except Exception as e:
                self.capture_exception(additional_context)
                if not silent:
                    raise e

        return wrapper

    def track_exception(self, func):
        """
        Decorator to be used for automatic exception capture, on HTTP 500 etc,
        where exception has already occurred
        """

        def wrapper(e):
            self.capture_exception()
            return func(e)

        return wrapper

    def get_exceptions(self, page_number=1):
        """
        Get list of exception objects from persistence store
        :param page_number: documents of a specific page
        :return: list of exception objects
        """
        if self.model:
            return self.model.get_exceptions_per_page(page_number=page_number).items
        raise ConfigError

    def get_exception(self, rhash):
        """
        Get a specific exception, can be used for some customization etc
        :param rhash:  hash of the exception
        :return:  exception object
        """
        if self.model:
            return self.model.get_entity(rhash)
        raise ConfigError

    def delete_exception(self, rhash):
        """
        Delete a specific exception from database
        :param rhash: hash of the exception
        :return:   whatever model returns
        """
        if self.model:
            return self.model.delete_entity(rhash)
        raise ConfigError

    def create_or_update_exception(self, rhash, host, path, method, request_data,
                                   exception_name, traceback):
        if self.model:
            return self.model.create_or_update_entity(rhash, host, path, method, request_data,
                                                      exception_name, traceback)
        raise ConfigError



# -*- coding: utf-8 -*-
#     Exception formatter defaults view for flask app
#
#     :copyright: 2020 Sonu Kumar
#     :license: BSD-3-Clause
#

import os

from flask import url_for, render_template, abort, redirect, request, blueprints

root_path = os.path.abspath(os.path.dirname(__file__))


class Views(object):
    def __init__(self, app, model, url_prefix, view_permission):
        self.model = model
        self.view_permission = view_permission
        blueprint = blueprints.Blueprint("app_error", 'app_error',
                                         root_path=root_path,
                                         template_folder="templates", url_prefix=url_prefix)

        @blueprint.route('/')
        def view_list():
            """
            List exceptions based on the page number
            :return: rendered template
            """
            if not view_permission(request):
                abort(401)
            title = "App Error"
            page = request.args.get('page', 1, type=int)
            error = False
            errors = model.get_exceptions_per_page(page_number=page)
            next_url = url_for('app_error.view_list', page=errors.next_num) \
                if errors.has_next else None
            prev_url = url_for('app_error.view_list', page=errors.prev_num) \
                if errors.has_prev else None
            return render_template('error_tracker/list.html', error=error, title=title, errors=errors,
                                   next_url=next_url, prev_url=prev_url)

        @blueprint.route('/<string:rhash>')
        def view_detail(rhash):
            """
            Display a specific exception having hash rhash
            :param rhash:  hash key of the exception
            :return: detailed view page
            """
            if not view_permission(request):
                abort(401)
            obj = model.get_entity(rhash)
            error = False
            if obj is None:
                abort(404)
            title = "%s : %s" % (obj.method, obj.path)

            return render_template('error_tracker/detail.html', error=error, title=title, obj=obj)

        @blueprint.route('/delete/<string:rhash>')
        def view_delete(rhash):
            """
            Delete a specific exceptions identified by rhash
            :param rhash: hash key of the exception
            :return: redirect back to the home page
            """
            if not view_permission(request):
                abort(401)
            model.delete_entity(rhash)
            return redirect(url_for('app_error.view_list'))

        app.register_blueprint(blueprint)



# -*- coding: utf-8 -*-
#
#     Exception formatter utils module
#
#     :copyright: 2020 Sonu Kumar
#     :license: BSD-3-Clause
#
import sys
import traceback
import warnings
from hashlib import sha256
import six

from error_tracker.libs.exception_formatter import format_exception
from error_tracker.libs.mixins import MaskingMixin


class Masking(MaskingMixin):
    """
    A simple function like class used for masking rule.
    """

    def __call__(self, key):
        if isinstance(key, six.string_types):
            tmp_key = key.lower()
            for k in self.mask_key_has:
                if k in tmp_key:
                    return True, "'%s'" % self.mask_with
        return False, None


class ConfigError(Exception):
    """
    A error class which will be raised by the app if it's not configure properly
    """
    pass


def get_exception_name(e):
    return str(e).replace("'>", "").replace("<class '", "").replace("<type 'exceptions.", "")


def get_context_detail(request, masking, context_builder,
                       additional_context):
    ty, val, tb = sys.exc_info()
    frames = traceback.format_exception(ty, val, tb)
    traceback_str = format_exception(tb, masking=masking)
    frame_str = ''.join(frames)
    rhash = sha256(str.encode(frame_str, "UTF-8")).hexdigest()
    request_data = context_builder.get_context(request, masking=masking,
                                               additional_context=additional_context)
    return ty, frames, frame_str, traceback_str, rhash, request_data


def get_notification_subject(notification_subject_prefix, method, url, exception):
    if notification_subject_prefix and method:
        subject = "[%s] [%s] %s" % (notification_subject_prefix, method, url)
    elif method:
        subject = "[%s] %s" % (method, url)
    else:
        subject = "[%s]" % notification_subject_prefix if notification_subject_prefix else ""
    if exception:
        subject = "%s %s" % (subject, exception)
    return subject


def get_class_from_path(module_path, super_class, raise_exception=True,
                        warning_message=None):
    """
    Try to import a class, that's subclass of super_class and it must have implemented all abstract methods
    :param module_path : path to the module
    :param super_class : super class
    :param raise_exception : in the case of import error, shallow it or re-raise
    :param warning_message: any warning message
    :return: imported class
    """
    from django.utils.module_loading import import_string
    try:
        cls = import_string(module_path)
        cls_fields = cls.__dict__.keys()
        matched = True
        if not issubclass(cls, super_class):
            return None
        fields = super_class.__dict__
        abstract_methods = None
        try:
            abstract_methods = cls.__abstractmethods__
        except AttributeError:
            pass
        if abstract_methods is not None:
            for method in abstract_methods:
                if method not in cls_fields:
                    matched = False
                    break
        else:
            for field, val in fields.items():
                if val is None:
                    continue
                if hasattr(val, '__isabstractmethod__'):
                    if field not in cls_fields:
                        matched = False
                        break
        if matched:
            return cls
    except ImportError as e:
        if raise_exception:
            six.reraise(ImportError, ImportError(e.message), sys.exc_info()[2])
        if warning_message:
            warnings.warn(warning_message)


def get_class_instance(module_path, mixin, default, message_prefix, *args):
    if module_path is not None:
        module = get_class_from_path(module_path, mixin,
                                     raise_exception=False,
                                     warning_message=message_prefix + " module " + module_path + " is not importable")
        if module is not None:
            return module(*args)
    if default is not None:
        message = "Default " + message_prefix + " module will be used"
        warnings.warn(message)
        return default(*args)


def get_context_dict(headers=None, context=None, form=None, args=None, masking=None):
    request_data = dict()
    form = form or {}
    headers = headers or {}
    args = args or {}
    context = context or {}
    if len(context) != 0:
        request_data['context'] = context
    if masking:
        for key in form:
            masked, value = masking(key)
            if masked:
                form[key] = value
        for key in headers:
            masked, value = masking(key)
            if masked:
                headers[key] = value
    if len(headers) != 0:
        request_data['headers'] = headers
    if len(args) != 0:
        request_data['args'] = args
    if len(form) != 0:
        request_data['form'] = form
    return request_data



# -*- coding: utf-8 -*-
#
#    Django error tracker default urls
#
#    :copyright: 2019 Sonu Kumar
#    :license: BSD-3-Clause
#

from django.conf.urls import url
from .views import detail, view_list, delete_exception

urlpatterns = [
    url(r'^$', view_list, name="view_errors"),
    url(r'^(?P<rhash>[\w-]+)/delete$', delete_exception, name='delete_error'),
    url(r'^(?P<rhash>[\w-]+)$', detail, name='view_error'),
]



# -*- coding: utf-8 -*-
#
#    Django error tracker template tags and filters
#
#    :copyright: 2019 Sonu Kumar
#    :license: BSD-3-Clause
#

from django import template

register = template.Library()


@register.filter
def replace_new_line_with_br(value):
    return value.replace("\n", "<br/>")



# -*- coding: utf-8 -*-
#     Utils modules for flask plugin
#
#     :copyright: 2020 Sonu Kumar
#     :license: BSD-3-Clause
#
from error_tracker import ContextBuilderMixin, ViewPermissionMixin


class DefaultFlaskContextBuilder(ContextBuilderMixin):
    """
    Default request builder, this records, form data, header and URL parameters and mask them if necessary
    """

    def get_context(self, request, masking=None, additional_context=None):
        request_data = dict()
        if additional_context is not None and len(additional_context) != 0:
            request_data['context'] = additional_context

        if request is not None:
            form = dict(request.form)
            headers = dict(request.headers)
            if masking:
                for key in form:
                    masked, value = masking(key)
                    if masked:
                        form[key] = value
                for key in headers:
                    masked, value = masking(key)
                    if masked:
                        headers[key] = value
            request_data.update({
                'headers': headers,
                'args': dict(request.args),
                'form': form
            })
        return str(request_data)


class DefaultFlaskViewPermission(ViewPermissionMixin):

    def __call__(self, request):
        return False


class configure_scope(object):
    """
    Use this class to work with context manager where more context can be added on the fly

    usage
    with configure_scope() as scope:
        scope.set_extra("id", 1234)
    """

    def __init__(self, error_manager, context=None, handle_exception=True):
        """
        Initialize class with error_manager instance.
        :param error_manager: AppErrorTracker instance
        :param handle_exception:  whether raised exception is ignored or not, post exception capture
        :param context: initial context detail, dictionary of key value pairs
        """
        self.context = context or {}
        self.error_manager = error_manager
        self.handle_exception = handle_exception

    def set_extra(self, key, value):
        self.context[key] = value

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if exc_type is not None:
            self.error_manager.capture_exception(additional_context=self.context)
        return self.handle_exception



# -*- coding: utf-8 -*-
#
#    Exception formatter mixin classes
#
#    :copyright: 2020 Sonu Kumar
#    :license: BSD-3-Clause
#

import abc


class MaskingMixin(object):
    __metaclass__ = abc.ABCMeta

    def __init__(self, mask_with, mask_key_has):
        """
        :param mask_with: masked value to be used for a given variable
        :param mask_key_has: tuple of strings to be used for checking masking rule
        """
        self.mask_key_has = mask_key_has
        self.mask_with = mask_with

    @abc.abstractmethod
    def __call__(self, key):
        raise NotImplementedError


class ModelMixin(object):
    """
    Base interface for data mode which can be used to store data in MongoDB/MySQL or any other data store.
    """
    hash = None
    host = None
    path = None
    method = None
    request_data = None
    exception_name = None
    traceback = None
    count = None
    created_on = None
    last_seen = None

    def __str__(self):
        return "'%s' '%s' %s" % (self.host, self.path, self.count)

    def __unicode__(self):
        return self.__str__()

    def __repr__(self):
        return "ModelMixin(%s)" % self.__str__()

    @classmethod
    @abc.abstractmethod
    def delete_entity(cls, rhash):
        """
        :param rhash: lookup key
        :return: None
        """
        raise NotImplementedError

    @classmethod
    @abc.abstractmethod
    def create_or_update_entity(cls, rhash, host, path, method, request_data,
                                exception_name, traceback):
        """
        :param rhash: Key of the db entry
        :param host: App host e.g. example.com
        :param path: request path
        :param method: request method (GET/POST/PUT etc)
        :param request_data: request form data
        :param exception_name: exception name
        :param traceback: Exception data captured
        :return:  error model object
        """
        raise NotImplementedError

    @classmethod
    @abc.abstractmethod
    def get_exceptions_per_page(cls, page_number=1):
        """
        An object having these properties,
        has_next, next_num, has_prev, prev_num and items
        on the returned object like SQLAlchemy's Pagination
        :return: a paginated object
        """
        raise NotImplementedError

    @classmethod
    @abc.abstractmethod
    def get_entity(cls, rhash):
        """
        :param rhash: key for lookup
        :return: Single entry of this class
        """
        raise NotImplementedError


class NotificationMixin(object):
    """
       A notification mixin class which can be used to notify any kind of notification.
    """
    __metaclass__ = abc.ABCMeta

    def __init__(self, *args, **kwargs):
        pass

    @abc.abstractmethod
    def notify(self, request, exception,
               email_subject=None,
               email_body=None,
               from_email=None,
               recipient_list=None):
        """
        :param request: current request context
        :param exception : exception model object
        :param email_subject: email subject
        :param email_body: email message object
        :param from_email: email sender as per config
        :param recipient_list: list of email ids for notification
        :return: None
        """
        raise NotImplementedError


class ContextBuilderMixin(object):
    """
       A context builder mixing that can be used to provide custom context details.
       This will provide entire details, that would be used for DB logging
       for usage see default context builder DefaultFlaskContextBuilder
    """

    __metaclass__ = abc.ABCMeta

    def __init__(self, *args, **kwargs):
        pass

    @abc.abstractmethod
    def get_context(self, request, masking=None,
                    additional_context=None):
        raise NotImplementedError


class TicketingMixin(object):
    """
    A mixing class that would be called with model object to raise the ticket.
    It can be used to directly create or update ticket in ticketing system
    """
    __metaclass__ = abc.ABCMeta

    def __init__(self, *args, **kwargs):
        pass

    @abc.abstractmethod
    def raise_ticket(self, exception, request=None):
        """
        :param exception: exception model object
        :param  request: current request
        :return: None
        """
        raise NotImplementedError


class ViewPermissionMixin(object):
    __metaclass__ = abc.ABCMeta

    @abc.abstractmethod
    def __call__(self, request):
        """
        Check whether user has the permission to view the request
        :param request: request object
        :return: True/False
        """
        raise NotImplementedError



import os
import re
from setuptools import setup


def fpath(name):
    return os.path.join(os.path.dirname(__file__), name)


def read(fname):
    return open(fpath(fname)).read()


def desc():
    info = read('README.rst')
    return info


file_text = read(fpath('error_tracker/__init__.py'))


def grep(attrname):
    pattern = r"{0}\W*=\W*'([^']+)'".format(attrname)
    strval, = re.findall(pattern, file_text)
    return strval


setup(
    name='error-tracker',
    version=grep('__version__'),
    url='https://github.com/sonus21/error-tracker/',
    license="BSD-3-Clause",
    author=grep('__author__'),
    author_email=grep('__email__'),
    description='Simple and Extensible Error Monitoring/Tracking framework for Python',
    keywords=['Flask', 'error-tracker', 'exception-tracking', 'exception-monitoring', "Django"],
    long_description=desc(),
    long_description_content_type='text/x-rst',
    packages=['error_tracker', ],
    include_package_data=True,
    python_requires=">=2.7, !=3.0.*, !=3.1.*, !=3.2.*, !=3.3.*, !=3.4.*",
    zip_safe=False,
    platforms='any',
    install_requires=[
        "six",
    ],
    extras_require={
        "Django": ["Django"],
        "DRF": ["djangorestframework"],
        "Flask": ["Flask", "Flask-SQLAlchemy"],
    },
    tests_require=[
        "Flask-Mail",
        'pyquery'
        "Django",
        "djangorestframework",
        "Flask",
        "Flask-SQLAlchemy"
    ],
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
        "License :: OSI Approved :: BSD License",
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ]
)



# -*- coding: utf-8 -*-
#
#    Django components
#
#    :copyright: 2020 Sonu Kumar
#    :license: BSD-3-Clause
#


from .utils import DefaultDjangoContextBuilder, DjangoNotification, DefaultDjangoViewPermission
from .settings import *

from .utils import DjangoNotification, DefaultDjangoContextBuilder
from error_tracker.libs.utils import Masking, get_class_from_path, get_class_instance
from error_tracker import ModelMixin, MaskingMixin, TicketingMixin, NotificationMixin, ContextBuilderMixin, \
    ViewPermissionMixin
from django.apps import apps as django_apps
import warnings


def get_exception_model():
    """
    Return the APP error model that is active in this project.
    """
    from .models import ErrorModel
    model_path = APP_ERROR_DB_MODEL
    if model_path is None:
        warnings.warn("APP_ERROR_DB_MODEL is not set using default model")
        return ErrorModel
    try:
        return django_apps.get_model(model_path, require_ready=False)
    except ValueError:
        model = get_class_from_path(model_path, ModelMixin, raise_exception=False,
                                    warning_message="Model " + model_path + " is not importable")
        if model is not None:
            return model
        warnings.warn("APP_ERROR_DB_MODEL must be of the form 'app_label.model_name'")
    except LookupError:
        model = get_class_from_path(model_path, ModelMixin, raise_exception=False,
                                    warning_message="Model " + model_path + " is not importable")
        if model is not None:
            return model
        raise warnings.warn(
            "APP_ERROR_DB_MODEL refers to model '%s' that has not been installed" % model_path
        )
    warnings.warn("APP_ERROR_DB_MODEL is set but not usable using default model")
    return ErrorModel


def get_masking_module():
    return get_class_instance(APP_ERROR_MASKING_MODULE, MaskingMixin, Masking, 'Masking', APP_ERROR_MASK_WITH,
                              APP_ERROR_MASKED_KEY_HAS)


def get_ticketing_module():
    return get_class_instance(APP_ERROR_TICKETING_MODULE, TicketingMixin, None, 'Ticketing')


def get_notification_module():
    if APP_ERROR_RECIPIENT_EMAIL and APP_ERROR_EMAIL_SENDER:
        return get_class_instance(APP_ERROR_NOTIFICATION_MODULE, NotificationMixin, DjangoNotification,
                                  "Notification")


def get_context_builder():
    return get_class_instance(APP_ERROR_CONTEXT_BUILDER_MODULE, ContextBuilderMixin,
                              DefaultDjangoContextBuilder, "ContextBuilder")


def get_view_permission():
    return get_class_instance(APP_ERROR_VIEW_PERMISSION, ViewPermissionMixin, DefaultDjangoViewPermission,
                              "ViewPermission")



# -*- coding: utf-8 -*-
#
#    Exception formatter that captures frame details in string format.
#
#    :copyright: 2020 Sonu Kumar
#    :license: BSD-3-Clause
#

import six
import itertools
import sys
import traceback

try:
    import builtins
except ImportError:
    import __builtin__ as builtins
try:
    from StringIO import StringIO
except ImportError:
    from io import StringIO

import re
import types


def convert_if_possible(x):
    try:
        from werkzeug.datastructures import ImmutableMultiDict
        if type(x) == ImmutableMultiDict:
            return "ImmutableMultiDict({%s})", x.to_dict(flat=False)
    except ImportError:
        pass
    try:
        from django.http import QueryDict
        if type(x) == QueryDict:
            return "QueryDict({%s})", x.dict()
    except ImportError:
        pass
    return None, None


def format_frame(x, max_elements, max_string, max_recursion, masking=None):
    """
    Return a formatted frame for storing in the database.
    :param x: frame key/value pair
    :param max_elements: Maximum number of elements to be formatted
    :param max_string: Maximum length for string data types in output
    :param max_recursion: Maximum recursion depth used in structure like dict
    :param masking: Used to mask a key/value
    :return: string of formatted value
    """

    def _per_element(i):
        return format_frame(i, max_elements, max_string, max_recursion - 1, masking=masking)

    def _per_dict_element(i):
        masked = False
        val = None
        if masking:
            masked, val = masking(i[0])
        return "%r : %s" % (i[0], val if masked else _per_element(i[1]))

    def _it_to_string(fmt, it, per_element=_per_element):
        if max_recursion <= 0:
            return fmt % "..."

        it = iter(it)

        s = ', '.join(per_element(i) for i in itertools.islice(it, max_elements))
        try:
            it.__next__()
        except AttributeError:
            try:
                next(it)
            except StopIteration:
                return fmt % s
        except StopIteration:
            return fmt % s

        # Add ellipsis indicating truncation.
        # Correctly handle the corner case of max_elements == 0.
        return fmt % (s + ", ..." if s else "...")

    if x is builtins.__dict__:
        return "<builtins>"
    if type(x) == dict:
        return _it_to_string("{%s}", sorted(x.items()), per_element=_per_dict_element)
    if type(x) == list:
        return _it_to_string("[%s]", x)
    if type(x) == tuple:
        return _it_to_string("(%s)" if len(x) != 1
                                       or max_recursion <= 0
                                       or max_elements <= 0
                             else "(%s,)", x)
    if type(x) == set:
        return _it_to_string("set([%s])", sorted(x))
    if type(x) == frozenset:
        return _it_to_string("frozenset([%s])", sorted(x))
    if isinstance(x, six.string_types) and max_string < len(x):
        return repr(x[:max_string] + "...")

    converted, x = convert_if_possible(x)
    if converted is not None:
        return _it_to_string(converted, sorted(x.items()),
                             per_element=_per_dict_element)
    try:
        if issubclass(x, dict):
            x = dict(x)
            return _it_to_string("Dict({%s})", sorted(x.items()),
                                 per_element=_per_dict_element)
    except TypeError:
        pass
    return repr(x)


def can_be_skipped(key, value):
    # Identifiers that are all uppercase are almost always constants.
    if re.match('[A-Z0-9_]+$', key):
        return True
    # dunder functions.
    if re.match('__.*__$', key):
        return True
    if callable(value):
        return True
    if isinstance(value, types.ModuleType):
        return True
    return False


def format_exception(tb, max_elements=1000,
                     max_string=10000, max_recursion=100,
                     masking=None):
    """
    :param tb: traceback
    :param max_elements:  Maximum number of elements to be printed
    :param max_string: Max string length in print
    :param max_recursion: Recursive printing in case of dict or other items
    :param masking: Masking rule for key/value pair
    :return: a formatted string

    Walk over all the frames and get the local variables from the frame and format them using format function and
    write to the stringIO based file.
    """
    stack = []
    t = tb
    while t:
        stack.append(t.tb_frame)
        t = t.tb_next
    buf = StringIO()
    w = buf.write
    # Go through each frame and format them to get final string
    for frame in stack:
        w('\n  File "%s", line %s, in %s\n' % (frame.f_code.co_filename,
                                               frame.f_lineno,
                                               frame.f_code.co_name))
        local_vars = frame.f_locals.items()
        local_vars = sorted(local_vars)
        for key, value in local_vars:
            if can_be_skipped(key, value):
                continue

            w("    %20s = " % key)
            masked = False
            if masking:
                masked, val = masking(key)
                if masked:
                    w(val)
            if not masked:
                try:
                    w(format_frame(value, max_elements, max_string, max_recursion,
                                   masking=masking))
                except Exception:
                    exc_class = sys.exc_info()[0]
                    w("<%s raised while printing value>" % exc_class)
            w("\n")
    w("\n")
    w(''.join(traceback.format_tb(tb)))
    op = buf.getvalue()
    buf.close()
    return op


def print_exception(masking=None, file=sys.stderr):
    """
    Print traceback in formatted format
    :return:  None
    """
    ty, val, tb = sys.exc_info()
    string = format_exception(tb, masking=masking)
    file.write(string)
